from ._src.dataclasses.dataclass import DataclassInstance, TDataclassInstance, dataclass
from ._src.dataclasses.helpers import as_shallow_dict, field

__all__ = ["DataclassInstance", "TDataclassInstance", "as_shallow_dict", "dataclass", "field"]
